@@ -1,23 +1,30 @@
+use std::convert::TryInto;
 use std::fs;
 use std::io;
 use std::io::prelude::*;
 use std::io::SeekFrom;
 use std::str;
+use std::marker::PhantomData;
 
 use log::warn;
 
 use lazy_static::lazy_static;
 use regex::Regex;
 
-use crate::peaks::{CentroidPeak, PeakCollection, PeakSet};
+use mzpeaks::{CentroidPeak, DeconvolutedPeak, PeakCollection, MZPeakSetType, MZLocated, IntensityMeasurement};
+use mzpeaks::peak::KnownCharge;
+
+use crate::params::{Param, ParamDescribed};
 use crate::spectrum::{
-    scan_properties, CentroidSpectrum, Precursor, RawSpectrum, SelectedIon, Spectrum,
-    SpectrumDescription,
-};
+    scan_properties, Precursor, SelectedIon, SpectrumDescription, SpectrumBehavior, PrecursorSelection};
+use crate::spectrum::spectrum::{
+    CentroidSpectrumType, MultiLayerSpectrum, CentroidPeakAdapting,
+    DeconvolutedPeakAdapting};
 
 use super::offset_index::OffsetIndex;
 use super::traits::{
     MZFileReader, RandomAccessScanIterator, ScanAccessError, ScanSource, SeekRead,
+    ScanWriter,
 };
 
 #[derive(PartialEq, Debug)]
@@ -41,74 +48,73 @@ pub enum MGFError {
 }
 
 #[derive(Debug, Clone)]
-struct SpectrumBuilder {
-    pub peaks: PeakSet,
+struct SpectrumBuilder<C: CentroidPeakAdapting = CentroidPeak, D: DeconvolutedPeakAdapting = DeconvolutedPeak> {
+    pub peaks: MZPeakSetType<C>,
     pub description: SpectrumDescription,
+    deconvoluted_type: PhantomData<D>
 }
 
-impl Default for SpectrumBuilder {
-    fn default() -> SpectrumBuilder {
+impl<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> Default for SpectrumBuilder<C, D> {
+    fn default() -> SpectrumBuilder<C, D> {
         SpectrumBuilder {
-            peaks: PeakSet::default(),
+            peaks: MZPeakSetType::<C>::default(),
             description: SpectrumDescription {
                 ms_level: 2,
                 signal_continuity: scan_properties::SignalContinuity::Centroid,
                 polarity: scan_properties::ScanPolarity::Unknown,
                 ..Default::default()
             },
+            deconvoluted_type: PhantomData
         }
     }
 }
 
-impl SpectrumBuilder {
-    pub fn into_spectrum(self, spectrum: &mut Spectrum) {
+impl<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> SpectrumBuilder<C, D> {
+    pub fn into_spectrum(self, spectrum: &mut MultiLayerSpectrum<C, D>) {
         spectrum.description = self.description;
         spectrum.peaks = Some(self.peaks);
     }
 }
 
-impl From<SpectrumBuilder> for Spectrum {
-    fn from(builder: SpectrumBuilder) -> Spectrum {
-        let mut spectrum = Spectrum::default();
+impl<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> From<SpectrumBuilder<C, D>> for MultiLayerSpectrum<C, D> {
+    fn from(builder: SpectrumBuilder<C, D>) -> MultiLayerSpectrum<C, D> {
+        let mut spectrum = MultiLayerSpectrum::<C, D>::default();
         builder.into_spectrum(&mut spectrum);
         spectrum
     }
 }
 
-macro_rules! impl_from_spectrum_builder_for_spec {
-    ($($t:ty), +) => {$(
-        impl From<SpectrumBuilder> for $t {
-            fn from(builder: SpectrumBuilder) -> $t {
-                let spec: Spectrum = builder.into();
-                spec.into()
-            }
-        }
-    )+};
+impl<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> From<SpectrumBuilder<C, D>> for CentroidSpectrumType<C> {
+    fn from(builder: SpectrumBuilder<C, D>) -> CentroidSpectrumType<C> {
+        let spec: MultiLayerSpectrum<C, D> = builder.into();
+        spec.try_into().unwrap()
+    }
 }
 
-impl_from_spectrum_builder_for_spec!(CentroidSpectrum, RawSpectrum);
-
 /// An MGF (Mascot Generic Format) file parser that supports iteration and random access.
 /// The parser produces [`Spectrum`] instances. These may be converted directly into [`CentroidSpectrum`]
 /// instances using [`Spectrum.into_centroid`] or the [`From`] trait which better represent
 /// the nature of this preprocessed data type.
-pub struct MGFReader<R: io::Read> {
+pub struct MGFReaderType<R: io::Read, C: CentroidPeakAdapting=CentroidPeak, D: DeconvolutedPeakAdapting=DeconvolutedPeak> {
     pub handle: io::BufReader<R>,
     pub state: MGFParserState,
     pub offset: usize,
     pub error: MGFError,
     pub index: OffsetIndex,
+    centroid_type: PhantomData<C>,
+    deconvoluted_type: PhantomData<D>
 }
 
-impl<R: io::Read> MGFReader<R> {
-    fn parse_peak_from_line(&mut self, line: &str) -> Option<CentroidPeak> {
+// A lazily created static regular expression to parse peak separators
+lazy_static! {
+    static ref PEAK_SEPERATOR: Regex = Regex::new(r"\t|\s+").unwrap();
+}
+
+impl<R: io::Read, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> MGFReaderType<R, C, D> {
+    fn parse_peak_from_line(&mut self, line: &str) -> Option<C> {
         let mut chars = line.chars();
         let first = chars.next().unwrap();
         if first.is_numeric() {
-            // A lazily created static regular expression to parse peak separators
-            lazy_static! {
-                static ref PEAK_SEPERATOR: Regex = Regex::new(r"\t|\s+").unwrap();
-            }
             let parts: Vec<&str> = PEAK_SEPERATOR.split(line).collect();
             let nparts = parts.len();
             if nparts < 2 {
@@ -121,7 +127,7 @@ impl<R: io::Read> MGFReader<R> {
                 mz,
                 intensity,
                 ..Default::default()
-            });
+            }.into());
         }
         None
     }
@@ -130,7 +136,7 @@ impl<R: io::Read> MGFReader<R> {
         &mut self,
         line: &str,
         description: &mut SpectrumDescription,
-        peaks: &mut PeakSet,
+        peaks: &mut MZPeakSetType<C>,
     ) -> bool {
         let peak_line = match self.parse_peak_from_line(line) {
             Some(peak) => {
@@ -178,9 +184,10 @@ impl<R: io::Read> MGFReader<R> {
                     });
                 }
                 &_ => {
-                    description
-                        .annotations
-                        .insert(String::from(key.to_lowercase()), String::from(value));
+                    description.add_param(Param::key_value(
+                        String::from(key.to_lowercase()),
+                        String::from(value),
+                    ));
                 }
             };
 
@@ -192,7 +199,7 @@ impl<R: io::Read> MGFReader<R> {
         }
     }
 
-    fn handle_peak(&mut self, line: &str, peaks: &mut PeakSet) -> bool {
+    fn handle_peak(&mut self, line: &str, peaks: &mut MZPeakSetType<C>) -> bool {
         let peak_line = match self.parse_peak_from_line(line) {
             Some(peak) => {
                 peaks.push(peak);
@@ -232,8 +239,8 @@ impl<R: io::Read> MGFReader<R> {
     }
 
     /// Read the next spectrum from the file, if there is one.
-    pub fn read_next(&mut self) -> Option<Spectrum> {
-        let mut builder = SpectrumBuilder::default();
+    pub fn read_next(&mut self) -> Option<MultiLayerSpectrum<C, D>> {
+        let mut builder = SpectrumBuilder::<C, D>::default();
         match self._parse_into(&mut builder) {
             Ok(offset) => {
                 if offset > 0 {
@@ -250,7 +257,7 @@ impl<R: io::Read> MGFReader<R> {
     }
 
     /// Read the next spectrum's contents directly into the passed struct.
-    fn _parse_into(&mut self, spectrum: &mut SpectrumBuilder) -> Result<usize, MGFError> {
+    fn _parse_into(&mut self, spectrum: &mut SpectrumBuilder<C, D>) -> Result<usize, MGFError> {
         let mut buffer = String::new();
         let mut work = true;
         let mut offset: usize = 0;
@@ -297,7 +304,7 @@ impl<R: io::Read> MGFReader<R> {
         }
         Ok(offset)
     }
-    pub fn read_into(&mut self, spectrum: &mut Spectrum) -> Result<usize, MGFError> {
+    pub fn read_into(&mut self, spectrum: &mut MultiLayerSpectrum<C, D>) -> Result<usize, MGFError> {
         let mut accumulator = SpectrumBuilder::default();
         match self._parse_into(&mut accumulator) {
             Ok(sz) => {
@@ -308,20 +315,22 @@ impl<R: io::Read> MGFReader<R> {
         }
     }
     /// Create a new, unindexed MGF parser
-    pub fn new(file: R) -> MGFReader<R> {
+    pub fn new(file: R) -> MGFReaderType<R, C, D> {
         let handle = io::BufReader::with_capacity(500, file);
-        MGFReader {
+        MGFReaderType {
             handle,
             state: MGFParserState::Start,
             offset: 0,
             error: MGFError::NoError,
             index: OffsetIndex::new("spectrum".to_owned()),
+            centroid_type: PhantomData,
+            deconvoluted_type: PhantomData
         }
     }
 }
 
-impl<R: io::Read> Iterator for MGFReader<R> {
-    type Item = Spectrum;
+impl<R: io::Read, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> Iterator for MGFReaderType<R, C, D> {
+    type Item = MultiLayerSpectrum<C, D>;
 
     /// Read the next spectrum from the file.
     fn next(&mut self) -> Option<Self::Item> {
@@ -329,10 +338,10 @@ impl<R: io::Read> Iterator for MGFReader<R> {
     }
 }
 
-impl<R: SeekRead> MGFReader<R> {
-    /// Construct a new MGFReader and build an offset index
+impl<R: SeekRead, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> MGFReaderType<R, C, D> {
+    /// Construct a new MGFReaderType and build an offset index
     /// using [`Self::build_index`]
-    pub fn new_indexed(file: R) -> MGFReader<R> {
+    pub fn new_indexed(file: R) -> MGFReaderType<R, C, D> {
         let mut reader = Self::new(file);
         reader.build_index();
         reader
@@ -395,11 +404,12 @@ impl<R: SeekRead> MGFReader<R> {
     }
 }
 
-impl<R: SeekRead> ScanSource<Spectrum> for MGFReader<R> {
+impl<R: SeekRead, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> ScanSource<C, D, MultiLayerSpectrum<C, D>> for MGFReaderType<R, C, D> {
     /// Retrieve a spectrum by it's native ID
-    fn get_spectrum_by_id(&mut self, id: &str) -> Option<Spectrum> {
+    fn get_spectrum_by_id(&mut self, id: &str) -> Option<MultiLayerSpectrum<C, D>> {
         let offset_ref = self.index.get(id);
         let offset = offset_ref.expect("Failed to retrieve offset");
+        let index = self.index.index_of(id).unwrap();
         let start = self
             .handle
             .stream_position()
@@ -409,11 +419,17 @@ impl<R: SeekRead> ScanSource<Spectrum> for MGFReader<R> {
         let result = self.read_next();
         self.seek(SeekFrom::Start(start))
             .expect("Failed to restore offset");
-        result
+        match result {
+            Some(mut scan) => {
+                scan.description.index = index;
+                Some(scan)
+            }
+            None => None,
+        }
     }
 
     /// Retrieve a spectrum by it's integer index
-    fn get_spectrum_by_index(&mut self, index: usize) -> Option<Spectrum> {
+    fn get_spectrum_by_index(&mut self, index: usize) -> Option<MultiLayerSpectrum<C, D>> {
         let (_id, offset) = self.index.get_index(index)?;
         let byte_offset = offset;
         let start = self
@@ -424,7 +440,13 @@ impl<R: SeekRead> ScanSource<Spectrum> for MGFReader<R> {
         let result = self.read_next();
         self.seek(SeekFrom::Start(start))
             .expect("Failed to restore offset");
-        result
+        match result {
+            Some(mut scan) => {
+                scan.description.index = index;
+                Some(scan)
+            }
+            None => None,
+        }
     }
 
     /// Return the data stream to the beginning
@@ -436,7 +458,7 @@ impl<R: SeekRead> ScanSource<Spectrum> for MGFReader<R> {
 
     fn get_index(&self) -> &OffsetIndex {
         if !self.index.init {
-            warn!("Attempting to use an uninitialized offset index on MGFReader")
+            warn!("Attempting to use an uninitialized offset index on MGFReaderType")
         }
         &self.index
     }
@@ -446,7 +468,7 @@ impl<R: SeekRead> ScanSource<Spectrum> for MGFReader<R> {
     }
 }
 
-impl<R: SeekRead> RandomAccessScanIterator<Spectrum> for MGFReader<R> {
+impl<R: SeekRead, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> RandomAccessScanIterator<C, D, MultiLayerSpectrum<C, D>> for MGFReaderType<R, C, D> {
     fn start_from_id(&mut self, id: &str) -> Result<&Self, ScanAccessError> {
         match self._offset_of_id(id) {
             Some(offset) => match self.seek(SeekFrom::Start(offset)) {
@@ -478,7 +500,7 @@ impl<R: SeekRead> RandomAccessScanIterator<Spectrum> for MGFReader<R> {
     }
 }
 
-impl MZFileReader<Spectrum> for MGFReader<fs::File> {
+impl<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> MZFileReader<C, D, MultiLayerSpectrum<C, D>> for MGFReaderType<fs::File, C, D> {
     fn open_file(source: fs::File) -> Self {
         Self::new(source)
     }
@@ -488,10 +510,144 @@ impl MZFileReader<Spectrum> for MGFReader<fs::File> {
     }
 }
 
+
+pub type MGFReader<R> = MGFReaderType<R, CentroidPeak, DeconvolutedPeak>;
+
+
+pub struct MGFWriterType<R: io::Write, C: CentroidPeakAdapting=CentroidPeak, D: DeconvolutedPeakAdapting=DeconvolutedPeak> {
+    pub handle: io::BufWriter<R>,
+    pub offset: usize,
+    centroid_type: PhantomData<C>,
+    deconvoluted_type: PhantomData<D>
+}
+
+
+impl<W: io::Write, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> MGFWriterType<W, C, D> {
+    pub fn new(file: W) -> MGFWriterType<W, C, D> {
+        let handle = io::BufWriter::with_capacity(500, file);
+        MGFWriterType {
+            handle,
+            offset: 0,
+            centroid_type: PhantomData,
+            deconvoluted_type: PhantomData
+        }
+    }
+
+    pub fn write_header(&mut self, spectrum: &MultiLayerSpectrum<C, D>) -> io::Result<usize> {
+        let desc = spectrum.description();
+        if desc.ms_level == 1 {
+            log::warn!("Attempted to write an MS1 spectrum to MGF, {}, skipping.", desc.id);
+            return Ok(0);
+        }
+        let mut count = self.handle.write(br#"BEGIN IONS
+TITLE="#)?;
+        count += self.handle.write(desc.id.as_bytes())?;
+        count += self.handle.write(b"\nRTINSECONDS=")?;
+        // let s = spectrum.start_time().to_string();
+        count += self.handle.write(spectrum.start_time().to_string().as_bytes())?;
+        count += self.handle.write(b"\n")?;
+        match &desc.precursor {
+            Some(precursor) => {
+                let ion = precursor.ion();
+                count += self.handle.write(b"PEPMASS=")?;
+                count += self.handle.write(ion.mz.to_string().as_bytes())?;
+                count += self.handle.write(b" ")?;
+                count += self.handle.write(ion.intensity.to_string().as_bytes())?;
+                match ion.charge {
+                    Some(charge) => {
+                        count += self.handle.write(b" ")?;
+                        count += self.handle.write(charge.to_string().as_bytes())?;
+                    },
+                    None => {}
+                }
+                count += self.handle.write(b"\n")?;
+                for param in precursor.params() {
+                    count += self.handle.write(param.name.to_uppercase().as_bytes())?;
+                    count += self.handle.write(param.value.as_bytes())?;
+                }
+            },
+            None => {}
+        }
+        for param in desc.params() {
+            count += self.handle.write(param.name.to_uppercase().as_bytes())?;
+            count += self.handle.write(param.value.as_bytes())?;
+        }
+
+        Ok(count)
+    }
+
+    pub fn write_deconvoluted_centroids(&mut self, spectrum: &MultiLayerSpectrum<C, D>) -> io::Result<usize> {
+        let mut count = 0;
+        match &spectrum.deconvoluted_peaks {
+            Some(centroids) => {
+                for peak in centroids.iter().map(|p|p.as_centroid()) {
+                    count += self.handle.write(
+                        peak.mz().to_string().as_bytes())?;
+                    count += self.handle.write(b" ")?;
+                    count += self.handle.write(peak.intensity().to_string().as_bytes())?;
+                    count += self.handle.write(b" ")?;
+                    count += self.handle.write(peak.charge().to_string().as_bytes())?;
+                    count += self.handle.write(b"\n")?;
+                }
+            },
+            None => {}
+        }
+        Ok(count)
+    }
+
+    pub fn write_centroids(&mut self, spectrum: &MultiLayerSpectrum<C, D>) -> io::Result<usize> {
+        let mut count = 0;
+        match &spectrum.peaks {
+            Some(centroids) => {
+                for peak in centroids {
+                    count += self.handle.write(peak.mz().to_string().as_bytes())?;
+                    count += self.handle.write(b" ")?;
+                    count += self.handle.write(peak.intensity().to_string().as_bytes())?;
+                    count += self.handle.write(b"\n")?;
+                }
+            },
+            None => {}
+        }
+        Ok(count)
+    }
+}
+
+impl<W: io::Write, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting>
+    ScanWriter<W, C, D> for MGFWriterType<W, C, D> {
+
+    fn write(&mut self, spectrum: &MultiLayerSpectrum<C, D>) -> io::Result<usize> {
+        let desc = spectrum.description();
+        if desc.ms_level == 1 {
+            log::warn!("Attempted to write an MS1 spectrum to MGF, {}, skipping.", desc.id);
+            return Ok(0);
+        }
+        let mut count = self.write_header(spectrum)?;
+        if matches!(&spectrum.deconvoluted_peaks, Some(_)) {
+            count += self.write_deconvoluted_centroids(spectrum)?;
+            count += self.handle.write(b"END IONS\n")?;
+        }
+        else if matches!(&spectrum.peaks, Some(_)) {
+            count += self.write_centroids(spectrum)?;
+            count += self.handle.write(b"END IONS\n")?;
+        } else {
+            count += self.handle.write(b"END IONS\n")?;
+        }
+        self.offset += count;
+        Ok(count)
+    }
+
+    fn flush(&mut self) -> io::Result<()> {
+        self.handle.flush()
+    }
+}
+
+
+pub type MGFWriter<W> = MGFWriterType<W, CentroidPeak, DeconvolutedPeak>;
+
+
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::spectrum::spectrum::SpectrumBehavior;
     use std::fs;
     use std::path;
 
@@ -499,7 +655,7 @@ mod test {
     fn test_reader() {
         let path = path::Path::new("./test/data/small.mgf");
         let file = fs::File::open(path).expect("Test file doesn't exist");
-        let reader = MGFReader::new(file);
+        let reader = MGFReaderType::<_, CentroidPeak, DeconvolutedPeak>::new(file);
         let mut ms1_count = 0;
         let mut msn_count = 0;
         for scan in reader {
@@ -518,7 +674,7 @@ mod test {
     fn test_reader_indexed() {
         let path = path::Path::new("./test/data/small.mgf");
         let file = fs::File::open(path).expect("Test file doesn't exist");
-        let mut reader = MGFReader::new_indexed(file);
+        let mut reader = MGFReaderType::<_, CentroidPeak, DeconvolutedPeak>::new_indexed(file);
 
         let n = reader.len();
         let mut ms1_count = 0;
@@ -536,4 +692,28 @@ mod test {
         assert_eq!(ms1_count, 0);
         assert_eq!(msn_count, 34);
     }
+
+    #[test]
+    fn test_writer() -> io::Result<()> {
+        let buff: Vec<u8> = Vec::new();
+        let inner_writer = io::Cursor::new(buff);
+        let mut writer = MGFWriter::new(inner_writer);
+
+        let path = path::Path::new("./test/data/small.mgf");
+        let file = fs::File::open(path).expect("Test file doesn't exist");
+        let reader = MGFReader::new(file);
+
+        let mut count = 0;
+        for scan in reader {
+            count += writer.write(&scan)?;
+        }
+        writer.flush()?;
+        let inner_writer = writer.handle.into_inner()?;
+        let buffer = inner_writer.get_ref();
+        let text = str::from_utf8(buffer).unwrap();
+        // Not including platform-specific line endings
+        assert_eq!(count, 548732);
+        assert_eq!(count, text.len());
+        Ok(())
+    }
 }
