@@ -31,20 +31,20 @@
 //! It also provides a sorted data structure for representing peak lists, [`PeakSet`]
 //! and a trait implementing the majority of the logic, [`PeakCollection`].
 pub mod io;
-pub mod mass_error;
 pub mod meta;
+#[macro_use]
 pub mod params;
-pub mod peaks;
+// pub mod peaks;
 pub mod spectrum;
 mod utils;
 
-pub use crate::peaks::coordinate::{CoordinateDimension, CoordinateLike, Mass, MZ};
-
-pub use crate::mass_error::MassErrorType;
-pub use crate::peaks::peak::{CentroidPeak, DeconvolutedPeak};
-pub use crate::peaks::{PeakCollection, PeakSet};
+pub use mzpeaks::MassErrorType;
+pub use mzpeaks::{CentroidPeak, DeconvolutedPeak};
+pub use mzpeaks::{PeakCollection, PeakSet};
 
 pub use crate::io::mgf::MGFReader;
 pub use crate::io::mzml::MzMLReader;
 
+pub use crate::params::{Param, ParamDescribed, ParamList};
+
 pub use crate::spectrum::{CentroidSpectrum, RawSpectrum, SpectrumBehavior};
