@@ -1,3 +1,4 @@
+use crate::impl_param_described;
 use crate::params::ParamList;
 
 #[derive(Debug)]
@@ -28,3 +29,5 @@ pub struct InstrumentConfiguration {
     pub software_reference: String,
     pub id: String,
 }
+
+impl_param_described!(InstrumentConfiguration, Component);
