@@ -1,6 +1,5 @@
-use std::collections::HashMap;
-
-use super::SpectrumBehavior;
+use super::spectrum::{SpectrumBehavior, CentroidPeakAdapting, DeconvolutedPeakAdapting};
+use crate::impl_param_described;
 use crate::io::traits::ScanSource;
 use crate::params;
 
@@ -116,10 +115,12 @@ impl Precursor {
     /// This is useful when examining the area *around* where the precursor
     /// ion was or to obtain a snapshot of the retention time when the spectrum
     /// was scheduled.
-    pub fn precursor_spectrum<R, S>(&self, source: &mut R) -> Option<S>
+    pub fn precursor_spectrum<C, D, S, R>(&self, source: &mut R) -> Option<S>
     where
-        R: ScanSource<S>,
-        S: SpectrumBehavior,
+        C: CentroidPeakAdapting,
+        D: DeconvolutedPeakAdapting,
+        S: SpectrumBehavior<C, D>,
+        R: ScanSource<C, D, S>,
     {
         source.get_spectrum_by_id(&self.precursor_id)
     }
@@ -127,15 +128,47 @@ impl Precursor {
     /// Given a ScanSource object, look up the product scan in it.
     /// This is rarely needed unless you have manually separated [`Precursor`]
     /// objects from their spectra.
-    pub fn product_spectrum<R, S>(&self, source: &mut R) -> Option<S>
+    pub fn product_spectrum<C, D, S, R>(&self, source: &mut R) -> Option<S>
     where
-        R: ScanSource<S>,
-        S: SpectrumBehavior,
+        C: CentroidPeakAdapting,
+        D: DeconvolutedPeakAdapting,
+        S: SpectrumBehavior<C, D>,
+        R: ScanSource<C, D, S>,
     {
         source.get_spectrum_by_id(&self.product_id)
     }
 }
 
+pub trait PrecursorSelection: params::ParamDescribed {
+    fn ion(&self) -> &SelectedIon;
+    fn isolation_window(&self) -> &IsolationWindow;
+    fn precursor_id(&self) -> &str;
+    fn product_id(&self) -> &str;
+    fn activation(&self) -> &Activation;
+}
+
+impl PrecursorSelection for Precursor {
+    fn ion(&self) -> &SelectedIon {
+        &self.ion
+    }
+
+    fn isolation_window(&self) -> &IsolationWindow {
+        &self.isolation_window
+    }
+
+    fn precursor_id(&self) -> &str {
+        &self.precursor_id
+    }
+
+    fn product_id(&self) -> &str {
+        &self.product_id
+    }
+
+    fn activation(&self) -> &Activation {
+        &self.activation
+    }
+}
+
 #[repr(i8)]
 #[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
 pub enum ScanPolarity {
@@ -154,7 +187,7 @@ impl Default for ScanPolarity {
 #[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
 pub enum SignalContinuity {
     Unknown = 0,
-    Centroid = 1,
+    Centroid = 3,
     Profile = 5,
 }
 
@@ -176,5 +209,13 @@ pub struct SpectrumDescription {
     pub params: params::ParamList,
     pub acquisition: Acquisition,
     pub precursor: Option<Precursor>,
-    pub annotations: HashMap<String, String>,
 }
+
+impl_param_described!(
+    Acquisition,
+    Activation,
+    Precursor,
+    SelectedIon,
+    ScanEvent,
+    SpectrumDescription
+);
