@@ -14,12 +14,31 @@ use log::warn;
 use base64;
 use flate2::write::ZlibDecoder;
 
+use mzpeaks::{
+    CentroidLike, CentroidPeak, DeconvolutedPeak, DeconvolutedPeakSet, MZPeakSetType,
+    MassErrorType, PeakCollection,
+};
+
 use crate::params::ParamList;
-use crate::peaks::{CentroidPeak, DeconvolutedPeak, PeakCollection, PeakSet, DeconvolutedPeakSet};
 use crate::utils::neutral_mass;
 
 type Bytes = Vec<u8>;
 
+pub fn to_bytes<T>(data: &[T]) -> Bytes {
+    let n = data.len();
+    let z = mem::size_of::<T>();
+    let m = n * z;
+    unsafe {
+        let byte_buffer = slice::from_raw_parts(
+            data.as_ptr() as *const u8,
+            m,
+        );
+        let mut result = Bytes::new();
+        result.copy_from_slice(byte_buffer);
+        return result
+    };
+}
+
 #[derive(Debug, Clone, Copy, PartialEq, Hash)]
 pub enum BinaryDataArrayType {
     Unknown,
@@ -165,6 +184,25 @@ impl<'transient, 'lifespan: 'transient> DataArray {
         }
     }
 
+    pub fn wrap(name: &ArrayType, dtype: BinaryDataArrayType, data: Bytes) -> DataArray {
+        DataArray {
+            dtype,
+            name: name.clone(),
+            data: data,
+            compression: BinaryCompressionType::Decoded,
+            ..Default::default()
+        }
+    }
+
+    pub fn update_buffer<T>(&mut self, data_buffer: &[T]) -> Result<usize, ArrayRetrievalError> {
+        if self.dtype.size_of() != mem::size_of::<T>() {
+            Err(ArrayRetrievalError::DataTypeSizeMismatch)
+        } else {
+            self.data = to_bytes(data_buffer);
+            Ok(self.data.len())
+        }
+    }
+
     fn decompres_zlib(&self, bytestring: &[u8]) -> Bytes {
         let result = Bytes::new();
         let mut decompressor = ZlibDecoder::new(result);
@@ -207,8 +245,26 @@ impl<'transient, 'lifespan: 'transient> DataArray {
         }
     }
 
+    pub fn decode_mut(&'transient mut self) -> Result<&'transient mut Bytes, ArrayRetrievalError> {
+        match self.compression {
+            BinaryCompressionType::Decoded => Ok(&mut self.data),
+            BinaryCompressionType::NoCompression => {
+                let bytestring = base64::decode(&self.data).expect("Failed to decode base64 array");
+                self.data = bytestring;
+                self.compression = BinaryCompressionType::Decoded;
+                Ok(&mut self.data)
+            }
+            BinaryCompressionType::Zlib => {
+                let bytestring = base64::decode(&self.data).expect("Failed to decode base64 array");
+                self.data = bytestring;
+                self.compression = BinaryCompressionType::Decoded;
+                Ok(&mut self.data)
+            }
+            _ => Err(ArrayRetrievalError::DecompressionError),
+        }
+    }
+
     pub fn coerce_from<T: Clone + Sized>(
-        &'lifespan self,
         buffer: Cow<'transient, Bytes>,
     ) -> Result<Cow<'transient, [T]>, ArrayRetrievalError> {
         let n = buffer.len();
@@ -225,11 +281,36 @@ impl<'transient, 'lifespan: 'transient> DataArray {
         }
     }
 
+    pub fn coerce_from_mut<T: Clone + Sized>(buffer: &mut [u8]) -> Result<&'transient mut [T], ArrayRetrievalError> {
+        let n = buffer.len();
+        let z = mem::size_of::<T>();
+        if n % z != 0 {
+            return Err(ArrayRetrievalError::DataTypeSizeMismatch);
+        }
+        let m = n / z;
+        unsafe {
+            Ok(slice::from_raw_parts_mut(
+                buffer.as_ptr() as *mut T,
+                m,
+            ))
+        }
+    }
+
+    pub fn coerce_mut<T: Clone + Sized>(&'lifespan mut self) -> Result<&'transient mut [T], ArrayRetrievalError> {
+        let view = match self.decode_mut() {
+            Ok(data) => data,
+            Err(err) => {
+                return Err(err)
+            }
+        };
+        Self::coerce_from_mut(view)
+    }
+
     pub fn coerce<T: Clone + Sized>(
         &'lifespan self,
     ) -> Result<Cow<'transient, [T]>, ArrayRetrievalError> {
         match self.decode() {
-            Ok(data) => self.coerce_from(data),
+            Ok(data) => Self::coerce_from(data),
             Err(err) => Err(err),
         }
     }
@@ -311,6 +392,69 @@ impl<'transient, 'lifespan: 'transient> DataArray {
             _ => Err(ArrayRetrievalError::DataTypeSizeMismatch),
         }
     }
+
+    pub fn to_i64(&'lifespan self) -> Result<Cow<'transient, [i64]>, ArrayRetrievalError> {
+        type D = i64;
+        match self.dtype {
+            BinaryDataArrayType::Float32 => {
+                type S = f32;
+                self.transmute::<S, D>()
+            }
+            BinaryDataArrayType::Float64 => {
+                type S = f64;
+                self.transmute::<S, D>()
+            }
+            BinaryDataArrayType::Int64 => self.coerce::<D>(),
+            BinaryDataArrayType::Int32 => {
+                type S = i32;
+                self.transmute::<S, D>()
+            }
+            _ => Err(ArrayRetrievalError::DataTypeSizeMismatch),
+        }
+    }
+
+    pub fn store_as(&mut self, dtype: BinaryDataArrayType) -> Result<usize, ArrayRetrievalError> {
+        if self.dtype == dtype {
+            return Ok(self.data.len());
+        }
+        let result = match dtype {
+            BinaryDataArrayType::Float32 => {
+                let view = match self.to_f32() {
+                    Ok(view) => view,
+                    Err(err) => {return Err(err)}
+                };
+                let recast = to_bytes(&view);
+                self.update_buffer(&recast)
+            },
+            BinaryDataArrayType::Float64 => {
+                let view = match self.to_f64() {
+                    Ok(view) => view,
+                    Err(err) => {return Err(err)}
+                };
+                let recast = to_bytes(&view);
+                self.update_buffer(&recast)
+            },
+            BinaryDataArrayType::Int32 => {
+                let view = match self.to_i32() {
+                    Ok(view) => view,
+                    Err(err) => {return Err(err)}
+                };
+                let recast = to_bytes(&view);
+                self.update_buffer(&recast)
+            },
+            BinaryDataArrayType::Int64 => {
+                let view = match self.to_i64() {
+                    Ok(view) => view,
+                    Err(err) => {return Err(err)}
+                };
+                let recast = to_bytes(&view);
+                self.update_buffer(&recast)
+            }
+            _ => {Ok(0)}
+        };
+        self.dtype = dtype;
+        result
+    }
 }
 
 #[derive(Debug, Default, Clone)]
@@ -329,10 +473,14 @@ impl<'transient, 'lifespan: 'transient> BinaryArrayMap {
         self.byte_buffer_map.insert(array.name.clone(), array);
     }
 
-    pub fn get(&self, array_type: &ArrayType) -> Option<&DataArray> {
+    pub fn get(&'transient self, array_type: &ArrayType) -> Option<&'transient DataArray> {
         self.byte_buffer_map.get(array_type)
     }
 
+    pub fn get_mut(&mut self, array_type: &ArrayType) -> Option<&mut DataArray> {
+        self.byte_buffer_map.get_mut(array_type)
+    }
+
     pub fn has_array(&self, array_type: &ArrayType) -> bool {
         self.byte_buffer_map.contains_key(array_type)
     }
@@ -341,7 +489,37 @@ impl<'transient, 'lifespan: 'transient> BinaryArrayMap {
         self.byte_buffer_map.clear();
     }
 
-    pub fn mzs(&'lifespan self) -> Cow<'transient, [f64]> {
+    pub fn search(
+        &self,
+        query: f64,
+        error_tolerance: f64,
+        error_type: MassErrorType,
+    ) -> Option<usize> {
+        let mzs = self.mzs();
+        let lower = error_type.lower_bound(query, error_tolerance);
+        match mzs[..].binary_search_by(|m| m.partial_cmp(&lower).unwrap()) {
+            Ok(i) => {
+                let mut best_error = error_type.call(query, mzs[i]).abs();
+                let mut best_index = i;
+                let mut index = i + 1;
+                while index < mzs.len() {
+                    let error = error_type.call(query, mzs[index]).abs();
+                    if error < best_error {
+                        best_index = index;
+                        best_error = error;
+                    }
+                    index += 1;
+                }
+                if best_error < error_tolerance {
+                    return Some(best_index);
+                }
+                None
+            }
+            Err(_err) => None,
+        }
+    }
+
+    pub fn mzs(&'transient self) -> Cow<'transient, [f64]> {
         let mz_array = self
             .get(&ArrayType::MZArray)
             .expect("Did not find m/z array")
@@ -350,7 +528,7 @@ impl<'transient, 'lifespan: 'transient> BinaryArrayMap {
         mz_array
     }
 
-    pub fn intensities(&'lifespan self) -> Cow<'transient, [f32]> {
+    pub fn intensities(&'transient self) -> Cow<'transient, [f32]> {
         let intensities = self
             .get(&ArrayType::IntensityArray)
             .expect("Did not find intensity array")
@@ -368,13 +546,13 @@ impl<'transient, 'lifespan: 'transient> BinaryArrayMap {
                     None
                 }
             },
-            None => None
+            None => None,
         }
     }
 }
 
-impl From<PeakSet> for BinaryArrayMap {
-    fn from(peaks: PeakSet) -> BinaryArrayMap {
+impl<C: CentroidLike> From<MZPeakSetType<C>> for BinaryArrayMap {
+    fn from(peaks: MZPeakSetType<C>) -> BinaryArrayMap {
         let mut arrays = BinaryArrayMap::new();
 
         let mut mz_array = DataArray::from_name_type_size(
@@ -393,8 +571,8 @@ impl From<PeakSet> for BinaryArrayMap {
         intensity_array.compression = BinaryCompressionType::Decoded;
 
         for p in peaks.iter() {
-            let mz: f64 = p.mz;
-            let inten: f32 = p.intensity;
+            let mz: f64 = p.coordinate();
+            let inten: f32 = p.intensity();
 
             let raw_bytes: [u8; mem::size_of::<f64>()] = unsafe { mem::transmute(mz) };
             mz_array.data.extend(raw_bytes);
@@ -409,59 +587,70 @@ impl From<PeakSet> for BinaryArrayMap {
     }
 }
 
-impl From<BinaryArrayMap> for PeakSet {
-    fn from(arrays: BinaryArrayMap) -> PeakSet {
+impl<C: CentroidLike + From<CentroidPeak>> From<BinaryArrayMap> for MZPeakSetType<C> {
+    fn from(arrays: BinaryArrayMap) -> MZPeakSetType<C> {
         let mz_array = arrays.mzs();
         let intensity_array = arrays.intensities();
         let mut peaks = Vec::with_capacity(mz_array.len());
 
         for (i, (mz, intensity)) in mz_array.iter().zip(intensity_array.iter()).enumerate() {
-            peaks.push(CentroidPeak {
-                mz: *mz,
-                intensity: *intensity,
-                index: i as u32,
-            })
+            peaks.push(
+                CentroidPeak {
+                    mz: *mz,
+                    intensity: *intensity,
+                    index: i as u32,
+                }
+                .into(),
+            )
         }
 
-        PeakSet::new(peaks)
+        MZPeakSetType::<C>::new(peaks)
     }
 }
 
-impl From<&BinaryArrayMap> for PeakSet {
-    fn from(arrays: &BinaryArrayMap) -> PeakSet {
+impl<C: CentroidLike + From<CentroidPeak>> From<&BinaryArrayMap> for MZPeakSetType<C> {
+    fn from(arrays: &BinaryArrayMap) -> MZPeakSetType<C> {
         let mz_array = arrays.mzs();
         let intensity_array = arrays.intensities();
         let mut peaks = Vec::with_capacity(mz_array.len());
 
         for (i, (mz, intensity)) in mz_array.iter().zip(intensity_array.iter()).enumerate() {
-            peaks.push(CentroidPeak {
-                mz: *mz,
-                intensity: *intensity,
-                index: i as u32,
-            })
+            peaks.push(
+                CentroidPeak {
+                    mz: *mz,
+                    intensity: *intensity,
+                    index: i as u32,
+                }
+                .into(),
+            )
         }
 
-        PeakSet::new(peaks)
+        MZPeakSetType::<C>::new(peaks)
     }
 }
 
-
 impl From<&BinaryArrayMap> for DeconvolutedPeakSet {
     fn from(arrays: &BinaryArrayMap) -> DeconvolutedPeakSet {
         let mz_array = arrays.mzs();
         let intensity_array = arrays.intensities();
-        let charge_array = arrays.charges().expect(
-            "Charge state array is required for deconvoluted peaks");
+        let charge_array = arrays
+            .charges()
+            .expect("Charge state array is required for deconvoluted peaks");
         let mut peaks = Vec::with_capacity(mz_array.len());
-        for (i, ((mz, intensity), charge)) in mz_array.iter().zip(intensity_array.iter()).zip(charge_array.iter()).enumerate() {
+        for (i, ((mz, intensity), charge)) in mz_array
+            .iter()
+            .zip(intensity_array.iter())
+            .zip(charge_array.iter())
+            .enumerate()
+        {
             peaks.push(DeconvolutedPeak {
                 neutral_mass: neutral_mass(*mz, *charge),
                 intensity: *intensity,
                 charge: *charge,
-                index: i as u32
+                index: i as u32,
             })
         }
 
         DeconvolutedPeakSet::new(peaks)
     }
-}
\ No newline at end of file
+}
