@@ -14,25 +14,46 @@
 //! These structures all implement the [`SpectrumBehavior`] trait
 use std::borrow;
 use std::convert::TryFrom;
-use crate::mass_error::MassErrorType;
-use crate::peaks::{CentroidPeak};
-use crate::peaks::{PeakCollection, PeakSet, DeconvolutedPeakSet};
+
+use mzpeaks::prelude::*;
+use mzpeaks::{
+    CentroidLike, DeconvolutedCentroidLike, DeconvolutedPeakSet, MZPeakSetType, MassPeakSetType,
+    PeakSet
+};
+use mzpeaks::{CentroidPeak, DeconvolutedPeak, MassErrorType};
+
+use mzsignal::peak_picker::{PeakPicker, PeakPickerError, PeakFitType};
+use mzsignal::denoise::{DenoisingError, denoise};
+
+
 use crate::spectrum::scan_properties::{
     Acquisition, Precursor, SignalContinuity, SpectrumDescription,
+    ScanPolarity
 };
-use crate::spectrum::signal::{BinaryArrayMap, ArrayType};
+use crate::spectrum::signal::{ArrayType, BinaryArrayMap, BinaryDataArrayType};
+
+pub trait CentroidPeakAdapting: CentroidLike + Default + From<CentroidPeak> {}
+impl<C: CentroidLike + Default + From<CentroidPeak>> CentroidPeakAdapting for C {}
+pub trait DeconvolutedPeakAdapting:
+    DeconvolutedCentroidLike + Default + From<DeconvolutedPeak>
+{
+}
+impl<D: DeconvolutedCentroidLike + Default + From<DeconvolutedPeak>> DeconvolutedPeakAdapting
+    for D
+{
+}
 
 #[derive(Debug)]
 /// An variant for dispatching to different strategies of computing
 /// common statistics of different levels of peak data.
-pub enum PeakDataLevel<'lifespan> {
+pub enum PeakDataLevel<'lifespan, C: CentroidLike, D: DeconvolutedCentroidLike> {
     Missing,
     RawData(&'lifespan BinaryArrayMap),
-    Centroid(&'lifespan PeakSet),
-    Deconvoluted(&'lifespan DeconvolutedPeakSet),
+    Centroid(&'lifespan MZPeakSetType<C>),
+    Deconvoluted(&'lifespan MassPeakSetType<D>),
 }
 
-impl<'lifespan> PeakDataLevel<'lifespan> {
+impl<'lifespan, C: CentroidLike, D: DeconvolutedCentroidLike> PeakDataLevel<'lifespan, C, D> {
     /// Compute the base peak of a spectrum
     pub fn base_peak(&self) -> (usize, f64, f32) {
         match self {
@@ -53,20 +74,20 @@ impl<'lifespan> PeakDataLevel<'lifespan> {
                 let result = peaks
                     .iter()
                     .enumerate()
-                    .max_by(|ia, ib| ia.1.intensity.partial_cmp(&ib.1.intensity).unwrap());
+                    .max_by(|ia, ib| ia.1.intensity().partial_cmp(&ib.1.intensity()).unwrap());
                 if let Some((i, peak)) = result {
-                    (i, peak.mz, peak.intensity)
+                    (i, peak.coordinate(), peak.intensity())
                 } else {
                     (0, 0.0, 0.0)
                 }
-            },
+            }
             PeakDataLevel::Deconvoluted(peaks) => {
                 let result = peaks
                     .iter()
                     .enumerate()
-                    .max_by(|ia, ib| ia.1.intensity.partial_cmp(&ib.1.intensity).unwrap());
+                    .max_by(|ia, ib| ia.1.intensity().partial_cmp(&ib.1.intensity()).unwrap());
                 if let Some((i, peak)) = result {
-                    (i, peak.neutral_mass, peak.intensity)
+                    (i, peak.coordinate(), peak.intensity())
                 } else {
                     (0, 0.0, 0.0)
                 }
@@ -81,8 +102,8 @@ impl<'lifespan> PeakDataLevel<'lifespan> {
                 let intensities = arrays.intensities();
                 intensities.iter().sum()
             }
-            PeakDataLevel::Centroid(peaks) => peaks.iter().map(|p| p.intensity).sum(),
-            PeakDataLevel::Deconvoluted(peaks) => peaks.iter().map(|p| p.intensity).sum()
+            PeakDataLevel::Centroid(peaks) => peaks.iter().map(|p| p.intensity()).sum(),
+            PeakDataLevel::Deconvoluted(peaks) => peaks.iter().map(|p| p.intensity()).sum(),
         }
     }
 
@@ -94,38 +115,19 @@ impl<'lifespan> PeakDataLevel<'lifespan> {
     ) -> Option<usize> {
         match self {
             PeakDataLevel::Missing => None,
-            PeakDataLevel::RawData(arrays) => {
-                let mzs = arrays.mzs();
-                let lower = error_type.lower_bound(query, error_tolerance);
-                match mzs[..].binary_search_by(|m| m.partial_cmp(&lower).unwrap()) {
-                    Ok(i) => {
-                        let mut best_error = error_type.call(query, mzs[i]).abs();
-                        let mut best_index = i;
-                        let mut index = i + 1;
-                        while index < mzs.len() {
-                            let error = error_type.call(query, mzs[index]).abs();
-                            if error < best_error {
-                                best_index = index;
-                                best_error = error;
-                            }
-                            index += 1;
-                        }
-                        if best_error < error_tolerance {
-                            return Some(best_index);
-                        }
-                        None
-                    }
-                    Err(_err) => None,
-                }
-            }
+            PeakDataLevel::RawData(arrays) => arrays.search(query, error_tolerance, error_type),
             PeakDataLevel::Centroid(peaks) => peaks.search(query, error_tolerance, error_type),
-            PeakDataLevel::Deconvoluted(peaks) => peaks.search(query, error_tolerance, error_type)
+            PeakDataLevel::Deconvoluted(peaks) => peaks.search(query, error_tolerance, error_type),
         }
     }
 }
 
 /// A trait for providing a uniform delegated access to spectrum metadata
-pub trait SpectrumBehavior {
+pub trait SpectrumBehavior<
+    C: CentroidLike = CentroidPeak,
+    D: DeconvolutedCentroidLike = DeconvolutedPeak,
+>
+{
     /// The method to access the spectrum description itself, which supplies
     /// the data for most other methods on this trait.
     fn description(&self) -> &SpectrumDescription;
@@ -136,6 +138,7 @@ pub trait SpectrumBehavior {
     }
 
     /// Access the precursor information, if it exists.
+    #[inline]
     fn precursor(&self) -> Option<&Precursor> {
         let desc = self.description();
         if let Some(precursor) = &desc.precursor {
@@ -147,6 +150,7 @@ pub trait SpectrumBehavior {
 
     /// A shortcut method to retrieve the scan start time
     /// of a spectrum.
+    #[inline]
     fn start_time(&self) -> f64 {
         let acq = self.acquisition();
         match acq.scans.first() {
@@ -156,29 +160,38 @@ pub trait SpectrumBehavior {
     }
 
     /// Access the MS exponentiation level
+    #[inline]
     fn ms_level(&self) -> u8 {
         self.description().ms_level
     }
 
     /// Access the native ID string for the spectrum
+    #[inline]
     fn id(&self) -> &str {
         &self.description().id
     }
 
     /// Access the index of the spectrum in the source file
+    #[inline]
     fn index(&self) -> usize {
         self.description().index
     }
 
     /// Access a description of how raw the signal is, whether a
     /// profile spectrum is available or only centroids are present.
+    #[inline]
     fn signal_continuity(&self) -> SignalContinuity {
         self.description().signal_continuity
     }
 
+    #[inline]
+    fn polarity(&self) -> ScanPolarity {
+        self.description().polarity
+    }
+
     /// Retrieve the most processed representation of the mass spectrum's
     /// signal
-    fn peaks(&'_ self) -> PeakDataLevel<'_>;
+    fn peaks(&'_ self) -> PeakDataLevel<'_, C, D>;
 }
 
 #[derive(Default, Debug, Clone)]
@@ -200,6 +213,13 @@ pub enum SpectrumConversionError {
     NoPeakData,
 }
 
+#[derive(Debug, Clone)]
+pub enum SpectrumProcessingError {
+    DenoisingError(DenoisingError),
+    PeakPickerError(PeakPickerError),
+    SpectrumConversionError(SpectrumConversionError)
+}
+
 impl<'transient, 'lifespan: 'transient> RawSpectrum {
     /// Convert a spectrum into a [`CentroidSpectrum`]
     pub fn into_centroid(self) -> Result<CentroidSpectrum, SpectrumConversionError> {
@@ -243,55 +263,96 @@ impl<'transient, 'lifespan: 'transient> RawSpectrum {
     }
 
     /// Convert a spectrum into a [`Spectrum`]
-    pub fn into_spectrum(self) -> Result<Spectrum, SpectrumConversionError> {
-        Ok(Spectrum {
+    pub fn into_spectrum<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting>(
+        self,
+    ) -> Result<MultiLayerSpectrum<C, D>, SpectrumConversionError> {
+        Ok(MultiLayerSpectrum::<C, D> {
             arrays: Some(self.arrays),
             description: self.description,
             ..Default::default()
         })
     }
+
+    pub fn denoise(&mut self, scale: f32) -> Result<(), SpectrumProcessingError> {
+        let mut intensities_copy = self.arrays.intensities().into_owned();
+        let mz_array = self.arrays.mzs();
+        match denoise(&mz_array, &mut intensities_copy, scale) {
+            Ok(_) => {
+                let view = self.arrays.get_mut(&ArrayType::IntensityArray).unwrap();
+                view.store_as(BinaryDataArrayType::Float32).expect("Failed to reformat intensity array");
+                view.update_buffer(&intensities_copy).expect("Failed to update intensity array buffer");
+                Ok(())
+            },
+            Err(err) => {
+                Err(SpectrumProcessingError::DenoisingError(err))
+            }
+        }
+    }
+
+
+    pub fn pick_peaks_with_into(self, peak_picker: &PeakPicker) -> Result<MultiLayerSpectrum<CentroidPeak, DeconvolutedPeak>, SpectrumProcessingError> {
+        let mut result = self.into_spectrum().unwrap();
+        match result.pick_peaks_with(peak_picker) {
+            Ok(_) => Ok(result),
+            Err(err) => Err(err)
+        }
+    }
+
+    pub fn pick_peaks_into(self, signal_to_noise_threshold: f32, fit_type: PeakFitType) -> Result<MultiLayerSpectrum<CentroidPeak, DeconvolutedPeak>, SpectrumProcessingError> {
+        let mut peak_picker = PeakPicker::default();
+        peak_picker.fit_type = fit_type;
+        peak_picker.signal_to_noise_threshold = signal_to_noise_threshold;
+        self.pick_peaks_with_into(&peak_picker)
+    }
 }
 
 impl<'lifespan> SpectrumBehavior for RawSpectrum {
+    #[inline]
     fn description(&self) -> &SpectrumDescription {
         &self.description
     }
 
-    fn peaks(&'_ self) -> PeakDataLevel<'_> {
+    fn peaks(&'_ self) -> PeakDataLevel<'_, CentroidPeak, DeconvolutedPeak> {
         PeakDataLevel::RawData(&self.arrays)
     }
 }
 
 #[derive(Default, Debug, Clone)]
 /// Represents a spectrum that has been centroided
-pub struct CentroidSpectrum {
+pub struct CentroidSpectrumType<C: CentroidLike + Default> {
     /// The spectrum metadata describing acquisition conditions and details.
     pub description: SpectrumDescription,
     /// The picked centroid peaks
-    pub peaks: PeakSet,
+    pub peaks: MZPeakSetType<C>,
 }
 
-impl<'lifespan> SpectrumBehavior for CentroidSpectrum {
+impl<'lifespan, C: CentroidLike + Default> SpectrumBehavior<C> for CentroidSpectrumType<C> {
+    #[inline]
     fn description(&self) -> &SpectrumDescription {
         &self.description
     }
 
-    fn peaks(&'_ self) -> PeakDataLevel<'_> {
+    fn peaks(&'_ self) -> PeakDataLevel<'_, C, DeconvolutedPeak> {
         PeakDataLevel::Centroid(&self.peaks)
     }
 }
 
-impl CentroidSpectrum {
+impl<C: CentroidLike + Default> CentroidSpectrumType<C> {
     /// Convert a spectrum into a [`Spectrum`]
-    pub fn into_spectrum(self) -> Result<Spectrum, SpectrumConversionError> {
-        Ok(Spectrum {
+    pub fn into_spectrum<D: DeconvolutedPeakAdapting>(
+        self,
+    ) -> Result<MultiLayerSpectrum<C, D>, SpectrumConversionError> {
+        let val = MultiLayerSpectrum::<C, D> {
             peaks: Some(self.peaks),
             description: self.description,
             ..Default::default()
-        })
+        };
+        Ok(val)
     }
 }
 
+pub type CentroidSpectrum = CentroidSpectrumType<CentroidPeak>;
+
 #[derive(Default, Debug, Clone)]
 struct DeconvolutedSpectrum {
     /// The spectrum metadata describing acquisition conditions and details.
@@ -301,11 +362,12 @@ struct DeconvolutedSpectrum {
 }
 
 impl<'lifespan> SpectrumBehavior for DeconvolutedSpectrum {
+    #[inline]
     fn description(&self) -> &SpectrumDescription {
         &self.description
     }
 
-    fn peaks(&'_ self) -> PeakDataLevel<'_> {
+    fn peaks(&'_ self) -> PeakDataLevel<'_, CentroidPeak, DeconvolutedPeak> {
         PeakDataLevel::Deconvoluted(&self.deconvoluted_peaks)
     }
 }
@@ -313,24 +375,30 @@ impl<'lifespan> SpectrumBehavior for DeconvolutedSpectrum {
 #[derive(Default, Debug, Clone)]
 /// Represent a spectrum with multiple layers of representation of the
 /// peak data.
-///
-/// **Not directly used at this time**
-pub struct Spectrum {
+pub struct MultiLayerSpectrum<C: CentroidLike + Default, D: DeconvolutedCentroidLike + Default> {
     /// The spectrum metadata describing acquisition conditions and details.
     pub description: SpectrumDescription,
+
     /// The (potentially absent) data arrays describing the m/z, intensity,
     /// and potentially other measured properties
     pub arrays: Option<BinaryArrayMap>,
+
     // The (potentially absent) centroid peaks
-    pub peaks: Option<PeakSet>,
+    pub peaks: Option<MZPeakSetType<C>>,
+
+    // The (potentially absent) deconvoluted peaks
+    pub deconvoluted_peaks: Option<MassPeakSetType<D>>,
 }
 
-impl<'lifespan> SpectrumBehavior for Spectrum {
+impl<'lifespan, C: CentroidLike + Default, D: DeconvolutedCentroidLike + Default>
+    SpectrumBehavior<C, D> for MultiLayerSpectrum<C, D>
+{
+    #[inline]
     fn description(&self) -> &SpectrumDescription {
         &self.description
     }
 
-    fn peaks(&'_ self) -> PeakDataLevel<'_> {
+    fn peaks(&'_ self) -> PeakDataLevel<'_, C, D> {
         if let Some(peaks) = &self.peaks {
             PeakDataLevel::Centroid(peaks)
         } else if let Some(arrays) = &self.arrays {
@@ -341,11 +409,11 @@ impl<'lifespan> SpectrumBehavior for Spectrum {
     }
 }
 
-impl Spectrum {
-    /// Convert a spectrum into a [`CentroidSpectrum`]
-    pub fn into_centroid(self) -> Result<CentroidSpectrum, SpectrumConversionError> {
+impl<'lifespan, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> MultiLayerSpectrum<C, D> {
+    /// Convert a spectrum into a [`CentroidSpectrumType<C>`]
+    pub fn into_centroid(self) -> Result<CentroidSpectrumType<C>, SpectrumConversionError> {
         if let Some(peaks) = self.peaks {
-            let mut result = CentroidSpectrum {
+            let mut result = CentroidSpectrumType::<C> {
                 peaks,
                 description: self.description,
             };
@@ -354,16 +422,16 @@ impl Spectrum {
         } else {
             if self.signal_continuity() == SignalContinuity::Centroid {
                 if let Some(arrays) = &self.arrays {
-                    let peaks = PeakSet::from(arrays);
-                    let mut centroid = CentroidSpectrum {
+                    let peaks = MZPeakSetType::<C>::from(arrays);
+                    let mut centroid = CentroidSpectrumType::<C> {
                         description: self.description,
                         peaks,
                     };
                     centroid.description.signal_continuity = SignalContinuity::Centroid;
                     return Ok(centroid);
                 } else {
-                    let mut result = CentroidSpectrum {
-                        peaks: PeakSet::empty(),
+                    let mut result = CentroidSpectrumType::<C> {
+                        peaks: MZPeakSetType::<C>::empty(),
                         description: self.description,
                     };
                     result.description.signal_continuity = SignalContinuity::Centroid;
@@ -398,22 +466,84 @@ impl Spectrum {
             })
         }
     }
-}
 
-impl From<Spectrum> for CentroidSpectrum {
-    fn from(spectrum: Spectrum) -> CentroidSpectrum {
-        spectrum.into_centroid().unwrap()
+    pub fn pick_peaks_with(&mut self, peak_picker: &PeakPicker) -> Result<(), SpectrumProcessingError> {
+        if let Some(arrays) = &self.arrays {
+            let mz_array = arrays.mzs();
+            let intensity_array = arrays.intensities();
+            let mut acc = Vec::new();
+            match peak_picker.discover_peaks(&mz_array, &intensity_array, &mut acc) {
+                Ok(_) => {
+                    let peaks: MZPeakSetType<C> = acc.into_iter().map(|p| C::from(p.into())).collect();
+                    self.peaks = Some(peaks);
+                    Ok(())
+                },
+                Err(err) => {
+                    Err(SpectrumProcessingError::PeakPickerError(err))
+                }
+            }
+        } else {
+            Err(SpectrumProcessingError::SpectrumConversionError(SpectrumConversionError::NoPeakData))
+        }
+    }
+
+    pub fn pick_peaks(&mut self, signal_to_noise_threshold: f32, fit_type: PeakFitType) -> Result<(), SpectrumProcessingError> {
+        let mut peak_picker = PeakPicker::default();
+        peak_picker.fit_type = fit_type;
+        peak_picker.signal_to_noise_threshold = signal_to_noise_threshold;
+        self.pick_peaks_with(&peak_picker)
+    }
+
+    pub fn denoise(&mut self, scale: f32) -> Result<(), SpectrumProcessingError> {
+        match &mut self.arrays {
+            Some(arrays) => {
+                let mut intensities_copy = arrays.intensities().into_owned();
+                let mz_array = arrays.mzs();
+                match denoise(&mz_array, &mut intensities_copy, scale) {
+                    Ok(_) => {
+                        let view = arrays.get_mut(&ArrayType::IntensityArray).unwrap();
+                        view.store_as(BinaryDataArrayType::Float32).expect("Failed to reformat intensity array");
+                        view.update_buffer(&intensities_copy).expect("Failed to update intensity array buffer");
+                        Ok(())
+                    },
+                    Err(err) => {
+                        Err(SpectrumProcessingError::DenoisingError(err))
+                    }
+                }
+            },
+            None => {
+                Err(SpectrumProcessingError::SpectrumConversionError(SpectrumConversionError::NoPeakData))
+            }
+        }
     }
 }
 
-impl From<CentroidSpectrum> for Spectrum {
-    fn from(spectrum: CentroidSpectrum) -> Spectrum {
+impl<'lifespan, C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting>
+    TryFrom<MultiLayerSpectrum<C, D>> for CentroidSpectrumType<C>
+{
+    type Error = SpectrumConversionError;
+
+    fn try_from(
+        spectrum: MultiLayerSpectrum<C, D>,
+    ) -> Result<CentroidSpectrumType<C>, Self::Error> {
+        spectrum.into_centroid()
+    }
+}
+
+pub type Spectrum = MultiLayerSpectrum<CentroidPeak, DeconvolutedPeak>;
+
+impl<C: CentroidPeakAdapting> From<CentroidSpectrumType<C>>
+    for MultiLayerSpectrum<C, DeconvolutedPeak>
+{
+    fn from(spectrum: CentroidSpectrumType<C>) -> MultiLayerSpectrum<C, DeconvolutedPeak> {
         spectrum.into_spectrum().unwrap()
     }
 }
 
-impl From<RawSpectrum> for Spectrum {
-    fn from(spectrum: RawSpectrum) -> Spectrum {
+impl<C: CentroidPeakAdapting, D: DeconvolutedPeakAdapting> From<RawSpectrum>
+    for MultiLayerSpectrum<C, D>
+{
+    fn from(spectrum: RawSpectrum) -> MultiLayerSpectrum<C, D> {
         spectrum.into_spectrum().unwrap()
     }
 }
@@ -424,9 +554,11 @@ impl From<Spectrum> for RawSpectrum {
     }
 }
 
-impl From<RawSpectrum> for CentroidSpectrum {
-    fn from(spectrum: RawSpectrum) -> CentroidSpectrum {
-        spectrum.into_centroid().unwrap()
+impl TryFrom<RawSpectrum> for CentroidSpectrum {
+    type Error = SpectrumConversionError;
+
+    fn try_from(spectrum: RawSpectrum) -> Result<CentroidSpectrum, Self::Error> {
+        spectrum.into_centroid()
     }
 }
 
@@ -442,11 +574,65 @@ impl TryFrom<Spectrum> for DeconvolutedSpectrum {
                 let peaks: DeconvolutedPeakSet = DeconvolutedPeakSet::from(arrays);
                 return Ok(DeconvolutedSpectrum {
                     description: spectrum.description,
-                    deconvoluted_peaks: peaks
-                })
+                    deconvoluted_peaks: peaks,
+                });
             }
-            return Err(Self::Error::NotDeconvoluted)
+            return Err(Self::Error::NotDeconvoluted);
+        }
+        return Err(Self::Error::NoPeakData);
+    }
+}
+
+
+#[cfg(test)]
+mod test {
+    use super::*;
+    use crate::io::prelude::*;
+    use crate::io::mzml::MzMLReader;
+
+    use mzsignal::plot::*;
+    use mzsignal::plot::{SVGBuilder, RED};
+    use mzsignal::arrayops::ArrayPair;
+
+    #[test]
+    fn test_profile_read() {
+        let mut reader = MzMLReader::open_path("./test/data/three_test_scans.mzML").expect("Failed to open test file");
+        let mut scan = reader.next().unwrap();
+        assert_eq!(scan.signal_continuity(), SignalContinuity::Profile);
+        assert_eq!(scan.ms_level(), 1);
+        assert_eq!(scan.polarity(), ScanPolarity::Positive);
+        assert!(matches!(scan.precursor(), None));
+
+
+
+        match scan.pick_peaks(1.0, PeakFitType::Quadratic) {
+            Err(err) => {
+                panic!("Should not have an error! {:?}", err);
+            },
+            Ok(_) => {}
+        }
+
+        if let Some(peaks) = &scan.peaks {
+            assert_eq!(peaks.len(), 2097);
+            let n = peaks.len();
+            for i in 1..n {
+                let diff = peaks[i].get_index() - peaks[i - 1].get_index();
+                assert_eq!(diff, 1);
+            }
+
+            let mut builder = SVGBuilder::default();
+            let arrays = &scan.arrays.unwrap();
+            let ser = ArrayPair::new(arrays.mzs(), arrays.intensities());
+            let mut ser2 = SpectrumSeries::from(peaks.iter());
+            ser2.color(RED.mix(1.0));
+            builder.path("./test/data/0.svg").size(1028, 512).add_series(
+                ser2
+            ).add_series(&ser).xlim(562f64, 565f64).draw().expect("Failed to draw");
+
+            peaks.has_peak(562.741, 3f64, MassErrorType::PPM).expect("Expected to find peak");
+            peaks.has_peak(563.240, 3f64, MassErrorType::PPM).expect("Expected to find peak");
+            let p = peaks.has_peak(563.739, 1f64, MassErrorType::PPM).expect("Expected to find peak");
+            assert!((p.mz() - 563.739).abs() < 1e-3)
         }
-        return Err(Self::Error::NoPeakData)
     }
 }
\ No newline at end of file
